"""Top-level package for phylotrackpy."""

__author__ = """Emily Dolson"""
__email__ = 'emilydolson@gmail.com'
__version__ = '0.1.5'

import phylotrackpy.systematics

__all__ = [
    "systematics",
]
